"""
winerp
~~~~~~

An extension module to facilitate creation of bot commands.

:copyright: (c) 2022-present BlackThunder
:license: MIT, see LICENSE for more details.
"""

from .client import Client
from .server import Server
from .lib.payload import winerpObject



import asyncio
from uuid import uuid4


class Payloads:
    '''
    Specifies all the payloads available in the system.
    This is a low level class for :class:`~winerp.lib.payload.PayloadTypes`.
    '''
    success = 0
    verification = 1
    request = 2
    response = 3
    error = 4
    ping = 5
    information = 6
    function_call = 7

class PayloadTypes:
    '''
    Specifies the type of message. Available Types:
        | ``success``: Successful authorization.
        | ``verification``: Verification request.
        | ``request``: Request for a route.
        | ``response``: Response to a request.
        | ``error``: Error response.
        | ``ping``: Ping message.
    '''
    def __init__(self, type: int) -> None:
        self._type = type
    
    def __repr__(self) -> str:
        return '<PayloadTypes: {}>'.format(self._type)
    
    @property
    def success(self) -> bool:
        '''
        :class:`bool`: Returns ``True`` if the message is a success message.
        '''
        return self._type == Payloads.success

    @property
    def verification(self) -> bool:
        '''
        :class:`bool`: Returns ``True`` if the message is a verification message.
        '''
        return self._type == Payloads.verification
    
    @property
    def request(self) -> bool:
        '''
        :class:`bool`: Returns ``True`` if the message is a request message.
        '''
        return self._type == Payloads.request
    
    @property
    def response(self) -> bool:
        '''
        :class:`bool`: Returns ``True`` if the message is a response message.
        '''
        return self._type == Payloads.response
    
    @property
    def error(self) -> bool:
        '''
        :class:`bool`: Returns ``True`` if the message is an error message.
        '''
        return self._type == Payloads.error

    @property
    def ping(self) -> bool:
        '''
        :class:`bool`: Returns ``True`` if the message is a ping message.
        '''
        return self._type == Payloads.ping

    @property
    def information(self) -> bool:
        '''
        :class:`bool`: Returns ``True`` if the message is an information message.
        '''
        return self._type == Payloads.information

    @property
    def function_call(self) -> bool:
        '''
        :class:`bool`: Returns ``True`` if the message is an information message.
        '''
        return self._type == Payloads.function_call



class MessagePayload:
    '''
    Represent IPC payload class.
    '''
    def __init__(self, **kwargs):
        self.id = kwargs.pop('id', None)
        self.type = kwargs.pop('type', None)
        self.route = kwargs.pop('route', None)
        self.traceback = kwargs.pop('traceback', None)
        self.data = kwargs.pop('data', {})
        self.uuid = kwargs.pop('uuid', None)
        self.destination = kwargs.pop('destination', None)
        self.pseudo_object = kwargs.pop('pseudo_object', None)
    

    def from_message(self, msg):
        '''
        Makes a payload from message. This is similar to cloning a message's ``dict`` to a new ``dict`` with same values.

        Parameters
        ----------
        msg: :class:`~winerp.lib.message.WsMessage`
            The message to clone
        
        Return Type
        -----------
        :class:`~winerp.lib.payload.MessagePayload`
        '''
        self.id = msg.id
        self.type = msg.type
        self.route = msg.route
        self.data = msg.data
        self.traceback = msg.traceback
        self.uuid = msg.uuid
        self.destination = msg.destination
        self.pseudo_object = msg.pseudo_object
        return self

    def to_dict(self) -> dict:
        '''
        Returns a ``dict`` representing the message.

        Return Type
        -----------
        :class:`dict`
        '''
        return {
            'id': self.id,
            'type': self.type,
            'route': self.route,
            'data': self.data,
            'traceback': self.traceback,
            'uuid': self.uuid,
            'destination': self.destination,
            'pseudo_object': self.pseudo_object
        }


class responseObject:
    def __init__(self, ipc_client, source, data):
        self.__ipc = ipc_client
        self.__name__ = data["__name__"]
        self.__uuid__ = data["__uuid__"]
        self.__source = source
        for each_attribute, attribute_value in data["__attr__"].items():
            self.__setattr__(each_attribute, attribute_value)

        for each_function, is_it_coro in data["__func__"].items():
            async def __async_fakeFunc(*args, **kwargs):
                return await self.__function_call(each_function, is_it_coro, *args, **kwargs)

            self.__setattr__(each_function, __async_fakeFunc)

    

    async def __function_call(self, function_name, is_it_coro, *args, **kwargs):
        return await self.__ipc.call_function(self.__source, self.__uuid__, function_name, *args, **kwargs)



class winerpObject:
    def __init__(self, object, required_functions = [], object_expiry=30, process_iters = True):
        """Creates a fake object which can be transferred to another client
        Whenever a fake object is sent to another client, the required functions are registered in the memory until the object expiry timeout.

        Args:
            object (class): The object you want to send to another client
            required_functions (list, optional): The class functions you want to execute on the client side. Defaults to [].
            object_expiry (int, optional): The time after which the object expires. Defaults to 30 seconds.
            process_iters (bool, optional): If set to True (default), all iterables with elements of datatype int, float, str, bool, & NoneType will be sent to the client.
        """
        self.object = object
        self.required_functions = required_functions
        self.object_expiry = object_expiry
        self.process_iters = process_iters
        self.uuid = str(uuid4())

    def __pythonic_object(self, var) -> bool:
        return isinstance(var, (int, float, str, bool, type(None)))

    def serialize_attributes(self):
        self.__serialized = {}
        self.functions = {}
        self.__serialized_functions = {}
        for attribute in self.object.__dir__():
            if attribute[0] == "_":
                continue
            
            try:
                attribute_value = getattr(self.object, attribute)
            except:
                continue
            if self.__pythonic_object(attribute_value):
                self.__serialized[attribute] = attribute_value
            elif callable(attribute_value):
                if attribute in self.required_functions:
                    self.functions[attribute] = attribute_value
                    self.__serialized_functions[attribute] = asyncio.iscoroutinefunction(attribute_value)
            elif isinstance(attribute_value, (list, set, tuple)):
                if self.process_iters:
                    if not any(not self.__pythonic_object(__elem) for __elem in attribute_value):
                        self.__serialized[attribute] = tuple(attribute_value) if isinstance(attribute_value, tuple) else attribute_value
            elif isinstance(attribute_value, dict):
                if self.process_iters:
                    if not any(
                        not self.__pythonic_object(key) or not self.__pythonic_object(value)
                        for key, value in attribute_value.items()):
                        self.__serialized[attribute] = attribute_value

            else:
                try:
                    if self.__pythonic_object(str(attribute_value.__str__())):
                        self.__serialized[attribute] = str(attribute_value)
                except:
                    ...
                
        return self.__serialized, self.functions

    def serialize(self):
        raw_object = self.serialize_attributes()
        return {
            "__name__": self.object.__class__.__name__,
            "__attr__": raw_object[0],
            "__func__": self.__serialized_functions,
            "__uuid__": self.uuid
        }



class UnauthorizedError(Exception):
    '''Raised when the client is tryint to make a request without being authorized by the server.'''
    pass


class UUIDNotFoundError(Exception):
    '''Raised when the specific UUID is not found in listeners.'''
    pass


class MissingUUIDError(Exception):
    '''Raised when the message does not have a ``uuid`` key.'''
    pass


class CheckFailureError(Exception):
    '''Raised when the check fails while the message is being dispatched.'''
    pass


class InvalidRouteType(Exception):
    '''Raised when the route type is not valid.'''
    pass


class ClientRuntimeError(Exception):
    '''Raised when the server returns back an error.'''
    pass

class ClientNotReadyError(Exception):
    '''Raised when the client is not ready to send requests'''
    pass




import argparse
from winerp.server import Server

def run():
    """ Module entry point"""

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-p",
        "--port",
        nargs='?',
        help="The port for winerp server",
        default=13254
    )

    parser.add_argument(
        "--version",
        action="store_true",
        help="print the version"
    )
    args = parser.parse_args()

    # Setup config from arguments
    port = args.port
    try:
        port = int(port)
    except:
        raise ValueError("port should be an integer between range 1-65535")
    
    if args.version:
        print('winerp version: unknown')
    else:
        print("Starting server at port: ", port)
        server = Server(port=port)
        server.start()


if __name__ == '__main__':
    run()



from .payload import PayloadTypes

class WsMessage:
    r"""
    Represents the message received from the server.
    """
    def __init__(self, message: dict):
        self._message = message
    
    def __repr__(self) -> str:
        return f'<winerp.WsMessage uuid={self.uuid} type={self.type.__repr__()}>'
    
    def __dict__(self) -> dict:
        return self.to_dict()
    
    @property
    def type(self) -> PayloadTypes:
        '''
        :class:`~winerp.lib.payload.PayloadTypes`: Returns the type of the message.
        '''
        return PayloadTypes(self._message["type"])
    
    @property
    def id(self) -> int:
        '''
        :class:`int`: Returns the id of the bot.
        '''
        return self._message.get("id")
    
    @property
    def destination(self) -> str:
        '''
        :class:`str`: Returns the destination of the message.
        '''
        return self._message.get("destination")
    
    @property
    def route(self) -> str:
        '''
        :class:`str`: Returns the route of the message.
        '''
        return self._message.get("route")
    
    @property
    def uuid(self) -> str:
        '''
        :class:`str`: Returns the unique id associated with this message.
        '''
        return self._message.get("uuid")
    
    @property
    def data(self) -> any:
        '''
        :class:`Any`: Returns the data associated with the message.
        '''
        return self._message.get("data")
    
    @property
    def error(self) -> str:
        '''
        :class:`str`: Returns the error associated with the message.
        '''
        return self._message.get("error")
    
    @error.setter
    def error(self, error: str):
        '''
        Sets the error associated with the message.
        '''
        self._message["error"] = error
    
    @property
    def traceback(self) -> str:
        '''
        :class:`str`: Returns the error associated with the message.
        '''
        return self._message.get("traceback")
    
    @traceback.setter
    def traceback(self, traceback: str):
        '''
        Sets the error associated with the message.
        '''
        self._message["traceback"] = traceback

    @property
    def pseudo_object(self) -> bool:
        '''
        :class:`str`: Returns the error associated with the message.
        '''
        return self._message.get("pseudo_object")
    

    @pseudo_object.setter
    def pseudo_object(self, pseudo_object: bool):
        '''
        Sets the error associated with the message.
        '''
        self._message["pseudo_object"] = pseudo_object

    def to_dict(self) -> dict:
        '''
        :class:`dict`: Returns the message as a `dict` type.
        '''
        return {
            'type': self.type,
            'id': self.id,
            'destination': self.destination,
            'route': self.route,
            'uuid': self.uuid,
            'data': self.data,
            'error': self.error,
            'traceback': self.traceback,
            'pseudo_object': self.pseudo_object
        }



import asyncio
from types import FunctionType
import websockets
import orjson
import logging
from .lib.message import WsMessage
from .lib.payload import Payloads, MessagePayload, winerpObject, responseObject
from .lib.errors import *
import uuid
import traceback
from typing import (
    Any,
    Callable,
    Coroutine,
    TypeVar,
    Union,
    Dict,
)

logger = logging.getLogger(__name__)
Coro = TypeVar('Coro', bound=Callable[..., Coroutine[Any, Any, Any]])

class Client:
    r"""Represents a winerp Client.
    This class is used to interact with the Server

    Parameters
    -----------
    local_name: :class:`str`
        The name which will be used to refer to this client.
        This should be unique to all the clients.
    port: Optional[:class:`int`]
        The port on which the server is running. Defaults to 13254.
    reconnect: Optional[:class:`bool`]
        If set to True, the client will automatically try to reconnect to the winerp server
        every 60 seconds (default). This option is set to True by default.
    """
    def __init__(
        self,
        local_name: str,
        port: int = 13254,
        reconnect: bool = True
    ):
        self.uri: str = f"ws://localhost:{port}"        
        self.local_name: str = local_name
        self.reconnect: bool = reconnect
        self.reconnect_threshold: int = 60
        self.max_data_size: float = 2 #MiB
        self.websocket = None
        self.__routes = {}
        self.__sub_routes = {}
        self.listeners = {}
        self.event_listeners: Dict[str, asyncio.Future] = {}
        self._authorized: bool = False
        self._on_hold = False
        self.events = [
            "on_winerp_connect",
            "on_winerp_ready",
            "on_winerp_disconnect",
            "on_winerp_request",
            "on_winerp_response",
            "on_winerp_information",
            "on_winerp_error"
        ]
    
    @property
    def authorized(self) -> bool:
        '''
        :class:`bool`: Returns if the client is authorized by the server.
        '''
        return self._authorized
    
    @property
    def on_hold(self) -> bool:
        '''
        :class:`bool`: Returns True if the client is on hold by the server. A client is put on hold if a client of same local name is already connected to the server.
        '''
        return self._on_hold

    async def send_message(self, data: Union[Any, WsMessage]):
        if not isinstance(data, WsMessage):
            data = data.__dict__
        logger.debug(data)
        await self.websocket.send(orjson.dumps(data).decode("utf-8"))
    
    def __send_message(self, data):
        asyncio.create_task(self.send_message(data))
    
    async def __verify_client(self):
        payload = MessagePayload(
            type = Payloads.verification,
            id = self.local_name,
            uuid = str(uuid.uuid4())
        )
        await self.send_message(payload)
        logger.info("Verification request sent")

    async def __connect(self) -> None:
        if self.websocket is None or self.websocket.closed:
            logger.info("Connecting to Websocket")
            self.websocket = await websockets.connect(
                self.uri, close_timeout=0, ping_interval=None, max_size=int(self.max_data_size*1048576)
            )
            self._authorized = False
            self._dispatch_event('winerp_connect')
            logger.info("Connected to Websocket")

    async def __reconnect_client(self) -> bool:
        while True:
            try:
                await self.__connect()
                await self.__verify_client()
                return True
            except:
                logger.debug(f"Failed to reconnect. Retrying in {self.reconnect_threshold}s")
                await asyncio.sleep(self.reconnect_threshold)
        
    async def start(self) -> None:
        '''|coro|

        Connects the client to the server.

        Raises
        -------
            ConnectionError
                If the websocket is already connected.
        
        Returns
        -------
            :class:`None`
        '''
        if self.websocket is None or self.websocket.closed:
            await self.__connect()
            await self.__verify_client()
            asyncio.create_task(self.__on_message())
        else:
            raise ConnectionError("Websocket is already connected!")

    def route(self, name: str = None):
        '''
        A decorator to register your route. The route name should be unique.

        Raises
        -------
            ValueError
                Route name already exists.
            InvalidRouteType
                The function passed is not a coro.
        '''
        def route_decorator(func):
            if (name is None and func.__name__ in self.__routes) or (name is not None and name in self.__routes):
                raise ValueError("Route name is already registered!")
            
            if not asyncio.iscoroutinefunction(func):
                raise InvalidRouteType("Route function must be a coro.")
            
            self.__routes[name or func.__name__] = func
            return func
            
        if isinstance(name, FunctionType):
            func = name
            name = name.__name__
            return route_decorator(func)
        else:
            return route_decorator

    async def __purge_sub_routes(self, timeout, uuid):
        await asyncio.sleep(timeout)
        del self.__sub_routes[uuid]
    def __register_object_funcs(self, winerp_object: winerpObject):
        self.__sub_routes[winerp_object.uuid] = {}
        for function_name, function_object in winerp_object.functions.items():
            self.__sub_routes[winerp_object.uuid][function_name] = function_object
        asyncio.create_task(self.__purge_sub_routes(winerp_object.object_expiry, winerp_object.uuid))

    async def ping(self, client = None, timeout: int = 60) -> bool:
        '''|coro|

        Pings the client and returns back if the ping was successful.

        Raises
        -------
            ClientNotReadyError
                The client is currently not ready to send or accept requests.
            UnauthorizedError
                The client isn't authorized by the server.
        
        Returns
        --------
            :class:`bool`
                If the ping is successful, it returns True.
        '''
        if self._on_hold or self.websocket is None or not self.websocket.open:
            raise ClientNotReadyError("The client is currently not ready to send or accept requests.")
        if not self._authorized:
            raise UnauthorizedError("Client is not authorized!")
        logger.debug("Pinging IPC Server")
        
        _uuid = str(uuid.uuid4())
        payload = MessagePayload(
            type = Payloads.ping,
            id = self.local_name,
            destination = client,
            uuid = _uuid
        )
        await self.send_message(payload)
        resp = await self.__get_response(_uuid, asyncio.get_event_loop(), timeout=timeout)
        return resp.get("success", False)

    async def call_function(self, destination, object_identifier, func_name, *args, **kwargs) -> bool:
        if self._on_hold or self.websocket is None or not self.websocket.open:
            raise ClientNotReadyError("The client is currently not ready to send or accept requests.")
        if not self._authorized:
            raise UnauthorizedError("Client is not authorized!")
        logger.debug("Calling a function IPC Server")
        
        _uuid = str(uuid.uuid4())
        payload = MessagePayload(
            type = Payloads.function_call,
            id = self.local_name,
            destination = destination,
            uuid = _uuid,
            data={
                "__uuid__": object_identifier,
                "__func__": func_name,
                "__args__": list(args),
                "__kwargs__": dict(kwargs)
            }
        )
        await self.send_message(payload)
        recv = await self.__get_response(_uuid, asyncio.get_event_loop(), timeout=30)
        return recv

    def __get_response(
        self,
        _uuid: str,
        loop: asyncio.AbstractEventLoop,
        timeout: int = 60
    ):
        future = loop.create_future()
        self.listeners[_uuid] = future
        return asyncio.wait_for(future, timeout, loop=loop)

    async def request(
        self,
        route: str,
        source: str,
        timeout: int = 60,
        **kwargs
    ) -> any:
        '''|coro|

        Requests the server for a response.
        Resolves when the response is received matching the UUID.

        Parameters
        -----------
        route: :class:`str`
            The route to request to.
        source: :class:`str`
            The destination
        timeout: :class:`int`
            Time to wait before raising :class:`~asyncio.TimeoutError`.

        Raises
        -------
            ClientNotReadyError
                The client is currently not ready to send or accept requests.
            UnauthorizedError
                The client isn't authorized by the server.
            ValueError:
                Missing either route or source or both.
            RuntimeError
                If the UUID is not found.
            asyncio.TimeoutError
                If the response is not received within the timeout.
        
        Returns
        --------
            :class:`any`
                The data associated with the message.
        '''
        if self.websocket is not None and self.websocket.open:
            if self._on_hold:
                raise ClientNotReadyError("The client is currently not ready to send or accept requests.")
            if not self._authorized:
                raise UnauthorizedError("Client is not authorized!")
            
            if not route or not source:
                raise ValueError("Missing required information for this request")

            logger.info("Requesting IPC Server for %r", route)
        
            _uuid = str(uuid.uuid4())
            payload = MessagePayload(
                type = Payloads.request,
                id = self.local_name,
                destination = source,
                route = route,
                data = kwargs,
                uuid = _uuid
            )

            await self.send_message(payload)
            recv = await self.__get_response(_uuid, asyncio.get_event_loop(), timeout=timeout)
            return recv
        
        else:
            raise ClientNotReadyError("The client has not been started or has disconnected")

    async def inform(
        self,
        data: any,
        destinations: list
    ):
        '''|coro|

        Sends data to other connected clients. There is no tracking of the data so there won't be any error
        if it doesn't reach its specified destination.

        The data is sent to all connected clients if the destinations list is empty.

        Parameters
        -----------
        data: :class:`Any`
            The data to redirect.
        destination: :class:`list`
            The list of destinations.

        Raises
        -------
            ClientNotReadyError
                The client is currently not ready to send or accept requests.
            UnauthorizedError
                The client isn't authorized by the server.
        
        Returns
        --------
            :class:`None`
        '''
        if self.websocket is not None and self.websocket.open:
            if self._on_hold:
                raise ClientNotReadyError("The client is currently not ready to send or accept requests.")
            if not self._authorized:
                raise UnauthorizedError("Client is not authorized!")

            logger.info("Informing IPC Server to redirect to routes %s" % destinations)
            if not isinstance(destinations, list):
                destinations = [destinations]

            payload = MessagePayload(
                type = Payloads.information,
                id = self.local_name,
                route = destinations,
                data = data,
            )

            await self.send_message(payload)
        else:
            raise ClientNotReadyError("The client has not been started or has disconnected")
    

    async def wait_until_ready(self):
        '''|coro|

        Waits until the client is ready to send or accept requests.        
        '''
        await self.wait_for('winerp_ready', None)

    async def wait_until_disconnected(self):
        '''|coro|
        
        Waits until the client is disconnected.
        '''
        await self.wait_for('winerp_disconnect', None)

    def wait_for(
        self,
        event: str,
        timeout: int = 60,
    ):
        '''|coro|

        Waits for a WebSocket event to be dispatched.

        The timeout parameter is passed onto asyncio.wait_for().
        By default, it does not timeout.

        In case the event returns multiple arguments, a tuple containing those arguments is returned instead.
        Please check the documentation for a list of events and their parameters.

        This function returns the **first event that meets the requirements.**

        Parameters
        -----------
        event: :class:`str`
            The event to wait for.
        timeout: Optional[:class:`int`]
            Time to wait before raising :class:`~asyncio.TimeoutError`. Defaults to 60.

        Raises
        -------
            asyncio.TimeoutError
                If the event is not received within the timeout.
        
        Returns
        --------
            :class:`Any`
                The payload for the event that meets the requirements.
        '''
        future = asyncio.get_event_loop().create_future()

        ev = event.lower()
        try:
            listeners = self.event_listeners[ev]
        except KeyError:
            listeners = []
            self.event_listeners[ev] = listeners

        listeners.append(future)
        return asyncio.wait_for(future, timeout)


    async def __on_message(self):
        logger.info("Listening to messages")
        while True:
            try:
                message = WsMessage(orjson.loads(await self.websocket.recv()))
            except websockets.exceptions.ConnectionClosedError:
                self._dispatch_event('winerp_disconnect')
                if self.reconnect:
                    if not await self.__reconnect_client():
                        break
                else:
                    break

            if message.type.success and not self._authorized:
                logger.info("Authorized Successfully")
                self._dispatch_event('winerp_ready')
                self._authorized = True
                self._on_hold = False

            elif message.type.ping:
                logger.debug("Received a ping from server")
                asyncio.create_task(self._dispatch(message))

            elif message.type.request:
                if message.route not in self.__routes:
                    logger.info("Failed to fulfill request, route not found")
                    payload = MessagePayload(
                        type = Payloads.error,
                        id = self.local_name,
                        data = "Route not found",
                        traceback = "Route not found",
                        destination = message.id,
                        uuid = message.uuid
                    )
                    self.__send_message(payload)
                    return
                logger.info("Fulfilling request @ route: %s" % message.route)
                asyncio.create_task(self._fulfill_request(message))
                self._dispatch_event('winerp_request')
            
            elif message.type.response:
                logger.info("Received a response from server @ uuid: %s" % message.uuid)
                asyncio.create_task(self._dispatch(message))
                self._dispatch_event('winerp_response')

            elif message.type.error:
                if message.data == "Already authorized.":
                    self._on_hold = True
                    logger.warn("Another client is already connected. Requests will be enabled when the other is disconnected.")
                else:
                    logger.debug("Failed to fulfill request: %s" % message.data)
                    self._dispatch_event('winerp_error', message.data)

                if message.uuid is not None:
                    asyncio.create_task(self._dispatch(message))

            elif message.type.information:
                if message.data:
                    logger.debug("Received an information bit from client: %s" % message.id)
                    self._dispatch_event('winerp_information', message.data, message.id)

            elif message.type.function_call:
                logger.debug("Received an object function call.")
                logger.debug(message.data)
                payload = MessagePayload(
                    type = Payloads.response,
                    id = self.local_name,
                    destination = message.id,
                    uuid = message.uuid
                )
                try:
                    called_function = self.__sub_routes[message.data["__uuid__"]][message.data["__func__"]]
                    asyncio.create_task(
                    self._fulfil_callback(
                        payload,
                        called_function,
                        *message.data["__args__"],
                        **message.data["__kwargs__"]
                    )
                )
                except KeyError:
                    payload = MessagePayload(
                        type = Payloads.error,
                        id = self.local_name,
                        data = "The called function has either expired or has never been registered",
                        traceback = "The called function has either expired or has never been registered",
                        destination = message.id,
                        uuid = message.uuid
                    )
                    self.__send_message(payload)

                

    def __parse_object(self, payload):
        payload.pseudo_object = True
        dummy_object = payload.data
        payload.data = dummy_object.serialize()
        self.__register_object_funcs(dummy_object)

    async def _fulfil_callback(self, payload, function, *args, **kwargs):
        try:
            payload.data = await function(*args, **kwargs)
            if not isinstance(payload.data, (int, float, str, bool, type(None), list, tuple, dict)):
                payload.data = winerpObject(payload.data)

            if type(payload.data) == winerpObject:
                self.__parse_object(payload)
            
            self.__send_message(payload)
        except Exception as error:
            logger.exception("Failed to run the registered method")
            self._dispatch_event('winerp_error', error)
            payload.type = Payloads.error
            payload.data = str(error)
            payload.traceback = ''.join(
                traceback.format_exception(
                    TypeError, error, error.__traceback__
                )
            )
        
    
    async def _fulfill_request(self, message: WsMessage):
        route = message.route
        func = self.__routes[route]
        data = message.data
        payload = MessagePayload().from_message(message)
        payload.type = Payloads.response
        payload.id = self.local_name


        try:
            payload.data = await func(message.id, **data)
            if type(payload.data) == winerpObject:
                self.__parse_object(payload)
        except Exception as error:
            logger.exception(error)
            self._dispatch_event('winerp_error', error)
            etype = type(error)
            trace = error.__traceback__
            lines = traceback.format_exception(etype, error, trace)
            traceback_text = ''.join(lines)

            payload.type = Payloads.error
            payload.data = str(error)
            payload.traceback = traceback_text
        finally:
            try:
                await self.send_message(payload)
            except TypeError as error:
                logger.exception("Failed to convert data to json")
                self._dispatch_event('winerp_error', error)
                payload.type = Payloads.error
                payload.data = str(error)
                payload.traceback = ''.join(
                    traceback.format_exception(
                        TypeError,
                        error,
                        error.__traceback__
                    )
                )
                self.__send_message(payload)

    async def _dispatch(self, msg: WsMessage):
        data = msg.data
        _uuid = msg.uuid
        if _uuid is None:
            raise MissingUUIDError('UUID is missing.')
        if _uuid not in self.listeners:
            raise UUIDNotFoundError(f"UUID {_uuid} not found in listeners.")
        
        future: asyncio.Future = self.listeners[_uuid]
        if not msg.type.error:
            if msg.pseudo_object:
                future.set_result(responseObject(self, msg.id, data))
            else:
                future.set_result(data)
        else:
            future.set_exception(
                ClientRuntimeError(msg.data)
            )

    def event(self, func: Coro, /) -> Coro:
        '''
        Registers a function for the event.

        The available events are:
            | ``on_winerp_connect``: The client has successfully connected to the server.
            | ``on_winerp_ready``: The client is ready to recieve and send requests.
            | ``on_winerp_disconnect``: The client has disconnected from the server.
            | ``on_winerp_request``: The server sent new request.
            | ``on_winerp_response``: The server sent back a response to a previous request.
            | ``on_winerp_information``: The server sent some data sourced by a client.
            | ``on_winerp_error``: An error occured during request processing.
        
        Raises
        -------
            NameError
                Invalid winerp event name.
            TypeError
                The event function is not a coro.
        '''
        if func.__name__ not in self.events:
            raise NameError("Invalid winerp event")
        
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("Event function must be a coro.")

        setattr(self, func.__name__, func)
        logger.debug(('%s has successfully been registered as an event', func.__name__))
        return func

    def _dispatch_event(self, event_name: str, *args, **kwargs):
        logger.debug('Event Dispatch -> %r', event_name)
        try:
            for future in self.event_listeners[event_name]:
                future.set_result(None)
                logger.debug('Event %r has been dispatched', event_name)
        except KeyError:
            ...

        try:
            coro = getattr(self, f'on_{event_name}')
        except AttributeError:
            pass
        else:
            self._schedule_event(coro, f'on_{event_name}', *args, **kwargs)

    
    def _schedule_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        *args: Any,
        **kwargs: Any,
    ) -> asyncio.Task:
        wrapped = self._run_event(coro, event_name, *args, **kwargs)
        # Schedules the task
        return asyncio.create_task(wrapped, name=f'winerp: {event_name}')
    

    async def _run_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        try:
            await coro(*args, **kwargs)
        except Exception:
            # TODO
            traceback.print_exc()



from websocket_server import WebsocketServer
from .lib.message import WsMessage
from .lib.payload import Payloads, MessagePayload
from .lib.errors import *
import orjson

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class Server:
    """
    Represents a winerp Server.
    This class is used as the central communication center for all connected clients.
    All requests and responses pass through the server

    If the library is installed using PyPi, you can also use terminal to start server using `winerp --port 1234`
    
    Parameters
    -----------
    port: Optional[:class:`int`]
        The port on which the server is running. Defaults to 13254.
    """
    def __init__(
        self,
        port: int = 13254
    ):
        self.websocket = WebsocketServer(host='127.0.0.1', port=port)
        self.websocket.set_fn_new_client(self.__on_client_connect)
        self.websocket.set_fn_message_received(self.__on_message)
        self.websocket.set_fn_client_left(self.__on_client_disconnect)
        self.active_clients = {}
        self.pending_verification = {}
        self.on_hold_connections = {}
    
    @property
    def client_count(self) -> int:
        '''
        :class:`int`: Returns the number of connected clients
        '''
        return len(self.active_clients)

    def __on_client_connect(self, client, _):
        logger.info("Client connected with id %s" % client['address'][1])
        self.pending_verification[client["address"][1]] = client

    def __on_client_disconnect(self, client, _):
        logger.info("Client disconnected with id %s" % client['address'][1])
        for cid, each_client in self.active_clients.items():
            if each_client["id"] == client["address"][1]:
                del self.active_clients[cid]
                if cid in self.on_hold_connections:
                    logger.info("On Hold Client moved to active client with connection id %s and local id %s" % (self.on_hold_connections[cid]['id'], cid))
                    self.active_clients[cid] = self.on_hold_connections[cid]
                    self.__send_message(
                        self.on_hold_connections[cid]["client"],
                        MessagePayload(type=Payloads.success, data="Authorized.")
                    )
                    del self.pending_verification[self.on_hold_connections[cid]["id"]]
                    del self.on_hold_connections[cid]
                return
        
        for cid, each_client in self.on_hold_connections.items():
            if each_client["id"] == client["address"][1]:
                del self.on_hold_connections[cid]
                return

        if client["address"][1] in self.pending_verification:
            del self.pending_verification[client["address"][1]]

        
    
    def __send_message(self, client, message):
        if not isinstance(message, dict):
            message = message.to_dict()
        self.websocket.send_message(
            client,
            orjson.dumps(message)
        )

    def __send_error(self, client, payload):
        if not isinstance(payload, dict):
            payload = payload.to_dict()

        self.websocket.send_message(
            client,
            orjson.dumps(payload)
        )

    def __on_message(self, client, _, msg):
        msg = WsMessage(orjson.loads(msg))
        payload = MessagePayload().from_message(msg)
        if msg.type.verification:
            if msg.id in self.active_clients:
                logger.info("Connection from duplicate client has benn put on hold connection id %s and local id %s" % (client['address'][1], msg.id))
                payload.uuid = None
                payload.type = Payloads.error
                payload.data = "Already authorized."
                payload.traceback = "Already authorized."
                self.__send_error(client, payload)
                self.on_hold_connections[msg.id] = {"client": client, "id": client["address"][1]}

            elif client["address"][1] in self.pending_verification:
                logger.info("Client verified with connection id %s and local id %s" % (client['address'][1], msg.id))
                self.active_clients[msg.id] = {"client": client, "id": client["address"][1]}
                del self.pending_verification[client["address"][1]]
                payload.type = Payloads.success
                payload.data = "Authorized."
                self.__send_message(client, payload)
        else:
            if client["address"][1] in self.pending_verification:
                logger.info('Unverified client tried to send message')
                payload.type = Payloads.error
                payload.data = "Not authorized."
                payload.traceback = "Not authorized."
                self.__send_error(client, payload)
                return
        
        if msg.type.information:
            logger.debug("Received Information Message from client %s" % client['address'][1])
            payload.type = Payloads.information
            if msg.route:
                for destination in msg.route:
                    if destination in self.active_clients:
                        payload.destination = destination
                        self.__send_message(self.active_clients[destination]["client"], payload)
            else:
                for client_id, client_obj in self.active_clients.items():
                    if client_id != payload.id:
                        payload.destination = client_id
                        self.__send_message(client_obj["client"], payload)

        if msg.type.ping:
            logger.debug("Received Ping Message from client %s" % client['address'][1])
            payload.type = Payloads.ping
            if (msg.destination is not None and msg.destination in self.active_clients) or msg.destination is None:
                payload.data = {"success": True}
            else:
                payload.data = {"success": False}
                
            self.__send_message(
                client,
                payload
            )

        if msg.type.request:
            logger.debug("Received Request Message from client %s" % client['address'][1])
            if msg.id == msg.destination:
                payload.type = Payloads.error
                payload.data = "Source and destination are the same."
                payload.traceback = "Source and destination are the same."
                self.__send_error(client, payload)

            elif msg.destination not in self.active_clients:
                payload.type = Payloads.error
                payload.data = "Destination not found."
                payload.traceback = "Destination not found."
                self.__send_error(client, payload)

            else:
                payload.type = Payloads.request
                payload.id = msg.destination
                payload.destination = msg.id
                self.__send_message(
                    self.active_clients[msg.destination]["client"],
                    payload
                )
                logger.debug("Request Message Forwarded to %s" % self.active_clients[msg.destination]["client"]['address'][1])

        if msg.type.response or msg.type.error or msg.type.function_call:
            logger.debug("Received Response Message from client %s" % client['address'][1])
            if msg.destination not in self.active_clients:
                payload.type = Payloads.error
                payload.data = "The data requester is no longer connected"
                payload.traceback = "The data requester is no longer connected"
                self.__send_error(client, payload)

            payload.type = Payloads.response if msg.type.response else (
                Payloads.function_call if msg.type.function_call else Payloads.error)

            self.__send_message(
                self.active_clients[msg.destination]["client"],
                payload
            )
            logger.debug("Response forwarded to %s" % self.active_clients[msg.destination]["client"]['address'][1])
            
            
    def start(self):
        '''
        Starts the server on the given port.
        '''
        logger.info("Started Websocket Server")
        self.websocket.run_forever()



from setuptools import setup
import pathlib

HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

extras = {
    'docs': [
        'sphinx==4.4.0',
        'sphinxcontrib_trio==1.1.2',
        'sphinxcontrib-websupport',
        'typing-extensions',
    ],
}

setup(
    name="winerp",
    version="1.4.0",
    description="Websocket based IPC for discord.py bots",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/BlackThunder01001/winerp",
    project_urls={
        "Bug Tracker": "https://github.com/BlackThunder01001/winerp/issues",
        "Documentation": "https://winerp.readthedocs.io/en/latest/",
    },
    author="BlackThunder",
    author_email="nouman0103@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
        "Typing :: Typed",
        
    ],
    packages=["winerp"],
    package_data={
     'winerp.lib': ['*'],
    },
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'winerp=winerp.__main__:run',
            ]
        },
    install_requires=["websockets", "websocket-server", "orjson"],
    extra_requires=extras,
    python_requires=">=3.7, !=3.10.*, !=3.11.*",
)
